# Ubuntu CI Services
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import socket
import time
import unittest

import mock
import webtest

from restish import resource
from restish.app import RestishApp

from bsbuilder import amqp_utils, utils


def _no_param_action():
    return utils.json_ok({'func': _no_param_action.__name__})


def _two_param_action(p1, p2):
    return utils.json_ok({'func': _two_param_action.__name__})


class TestGetResource(unittest.TestCase):
    '''Test to ensure the http_get_resource works properly'''

    def setUp(self):
        super(TestGetResource, self).setUp()

        class TR(resource.Resource):
            @resource.child()
            def empty(self, request, segments):
                return utils.http_get_resource(_no_param_action)

            @resource.child()
            def two_param_get(self, request, segments):
                return utils.http_get_resource(_two_param_action)

        app = RestishApp(TR())
        self.app = webtest.TestApp(app, relative_to='.')

    def testGetNoParam(self):
        resp = self.app.get('/empty', status=200)
        data = json.loads(resp.body)
        self.assertEqual(data['func'], _no_param_action.__name__)

    def testGetTwoParam(self):
        resp = self.app.get('/two_param_get', {'p1': 1, 'p2': 2}, status=200)
        data = json.loads(resp.body)
        self.assertEqual(data['func'], _two_param_action.__name__)

        # ensure a missing parameter returns a 400 error
        resp = self.app.get('/two_param_get', {'p1': 1}, status=400)
        self.assertTrue('p2' in resp.body)

    def testLeafEnforced(self):
        '''ensure we only respond to the leaf and not additional segments'''
        self.app.get('/empty/', status=404)

    def testAllowedMethods(self):
        self.app.post_json('/empty', {'key': 'val'}, status=405)


class TestPostResource(unittest.TestCase):
    '''Test to ensure the http_post_resource works properly'''

    def setUp(self):
        super(TestPostResource, self).setUp()

        class TR(resource.Resource):
            @resource.child()
            def empty(self, request, segments):
                return utils.http_post_resource(_no_param_action)

            @resource.child()
            def two_param_get(self, request, segments):
                return utils.http_post_resource(_two_param_action)

        app = RestishApp(TR())
        self.app = webtest.TestApp(app, relative_to='.')

    def testPostNoParam(self):
        resp = self.app.post_json('/empty', {}, status=200)
        data = json.loads(resp.body)
        self.assertEqual(data['func'], _no_param_action.__name__)

    def testPostTwoParam(self):
        resp = self.app.post_json('/two_param_get', {'p1': 1, 'p2': 2})
        data = json.loads(resp.body)
        self.assertEqual(data['func'], _two_param_action.__name__)

        # ensure a missing parameter returns a 400 error
        resp = self.app.post_json('/two_param_get', {'p1': 1}, status=400)
        self.assertTrue('p2' in resp.body)

    def testLeafEnforced(self):
        '''ensure we only respond to the leaf and not additional segments'''
        self.app.post_json('/empty/', {}, status=404)

    def testAllowedMethods(self):
        self.app.get('/empty', {'p1': 1, 'p2': 2}, status=405)


class TestAMQP(unittest.TestCase):
    @mock.patch('bsbuilder.amqp_utils.connection')
    @mock.patch('bsbuilder.amqp_utils.get_config')
    def testConnectFailed(self, get_config, connect):
        '''Ensure a failed queue connection returns an HTTP 503 error'''
        get_config.return_value = mock.Mock()
        error = 'mocked test exception'
        connect.side_effect = RuntimeError(error)
        r = amqp_utils.send('fake_queue', 'fake_message')
        self.assertIsNotNone(r)
        self.assertEqual(error, r)

    @mock.patch('bsbuilder.amqp_utils.connection')
    @mock.patch('bsbuilder.amqp_utils.get_config')
    def testSent(self, get_config, connect):
        '''Test a successful send returns nothing

        There's not much you can test in isolation here, but this gives
        a bit of a sanity check.
        '''
        get_config.return_value = mock.Mock()
        connect.return_value = mock.Mock()
        r = amqp_utils.send('fake_queue', 'fake_message')
        self.assertIsNone(r)

    @mock.patch('bsbuilder.amqp_utils._run_forever')
    @mock.patch('bsbuilder.amqp_utils.connection')
    def testProcessQueue(self, connection, run_forever):
        '''Ensure we close the connection if something fails'''
        conn = mock.Mock()
        connection.return_value = conn
        run_forever.side_effect = RuntimeError
        with self.assertRaises(RuntimeError):
            amqp_utils.process_queue(None, None, None)
        self.assertTrue(conn.close.called)

    def testRunForever(self):
        '''Ensure this times out after the right amount of time'''
        channel = mock.Mock()
        callback = mock.Mock()
        #wait = mock.Mock()
        #wait.side_effect = socket.error
        channel.wait.side_effect = socket.error()
        start = time.time()
        retry_period = 2
        amqp_utils._run_forever(channel, 'foo', callback, retry_period)
        # see if it was within 1/10 of a second of the retry_period
        self.assertAlmostEqual(start + retry_period, time.time(), places=1)
