# Ubuntu CI Services
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import logging
import json

from restish import http, resource

log = logging.getLogger(__name__)


def json_ok(data):
    '''Simple wrapper to return JSON data as an OK response'''
    data = json.dumps(data) + '\n'  # \n is handy when using curl from CLI
    return http.ok([('Content-Type', 'application/json')], data)


class _BaseResource(resource.Resource):
    '''Creates a resource that validates arguments by introspcection.

    This class makes it easy to supply a single function with parameters
    to handle HTTP operations. The GET/POST parameters will be compared to
    the callback function's parameter list to validate it can be called.
    '''
    def __init__(self, callback_func, allowed):
        super(_BaseResource, self).__init__()
        self._cb = callback_func
        self._allowed = allowed

    def _handle(self, req_args):
        args, vargs, keywords, defaults = inspect.getargspec(self._cb)
        missing = set(args) - set(req_args.keys())
        unknown = set(req_args.keys()) - set(args)
        if unknown:
            log.warning('Unsupported arguments: %r' % unknown)
        if missing:
            body = 'Missing required parameters\n %s\n' % '\n '.join(missing)
            return http.bad_request([('Content-Type', 'text/plain')], body)

        params = {k: req_args[k] for k in args}
        return self._cb(**params)

    @resource.GET()
    def _do_get(self, request):
        if resource.GET not in self._allowed:
            return http.method_not_allowed([request.method])
        return self._handle(request.GET)

    @resource.POST(accept='json')
    def _do_post(self, request):
        if resource.POST not in self._allowed:
            return http.method_not_allowed([request.method])
        params = json.loads(request.body)
        resp = self._handle(params)
        if not resp:
            resp = http.Response('204 No Content', [], None)
        return resp


def http_get_resource(callback_func):
    return _BaseResource(callback_func, (resource.GET,))


def http_post_resource(callback_func):
    return _BaseResource(callback_func, (resource.POST,))
