# Ubuntu CI Services
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import json
import logging
import urllib2

from django.conf import settings
from django.db import models


class LaunchpadError(Exception):
    '''Raised when an error occurrs interacting with Launchpad.'''


class PPA(models.Model):
    name = models.CharField(max_length=4096, unique=True)

    reserved = models.BooleanField(default=False)
    lockname = models.CharField(max_length=4096, blank=True)
    timestamp = models.DateTimeField(auto_now_add=True)

    def __unicode__(self):
        reserved = ''
        if self.reserved:
            reserved = ' (reserved)'
            if self.lockname:
                reserved = '%s - %s' % (reserved, self.lockname)
        return '%s %s' % (self.name, reserved)

    def clean_artifacts(self):
        """Cleans up all the artifacts residing in the ppa's."""
        raise LaunchpadError('TODO how to clean via LP')

    @staticmethod
    def _merge_ppas(lp_ppas):
        model_ppas = set([x.name for x in PPA.objects.all()])
        add = lp_ppas - model_ppas
        remove = model_ppas - lp_ppas

        logging.info('adding new ppas: %r', add)
        for x in add:
            PPA.objects.create(name=x)

        logging.info('removing ppas: %r', add)
        for x in remove:
            ppa = PPA.objects.get(name=x)
            if ppa.reserved:
                # log something, but we may as well delete. because if the
                # ppa doesn't exist in LP we are already in a catastrophic
                # situation and shouldn't let any other user re-use this
                logging.error('ppa %s appears to be in use', x)
            ppa.delete()

    @staticmethod
    def populate_from_launchpad():
        '''Get list of PPAs from launchpad and ensure they exist here.'''
        url = 'https://api.launchpad.net/1.0/~%s/ppas'
        url = url % settings.LAUNCHPAD_PPA_USER
        logging.info('populating ppa model from: %s', url)
        try:
            resp = urllib2.urlopen(url)
            resp = resp.read()
            data = json.loads(resp)
            # TODO test if this is needed or if there is some type
            # of looping logic needed
            assert data['total_size'] == len(data['entries'])
            ppas = ['ppa:%s/%s' % (settings.LAUNCHPAD_PPA_USER, x['name'])
                    for x in data['entries']]
            PPA._merge_ppas(set(ppas))

        except urllib2.URLError as e:
            logging.exception('Could not populate PPAs from launchpad')
            raise LaunchpadError(e)

    @staticmethod
    def reserve(lockname='', clean=False):
        # "select_for_update" provides proper db-level locking so we
        # don't accidentally have 2 concurrent requests find and reserve the
        # same PPA
        ppa = None
        qs = PPA.objects.select_for_update().filter(reserved=False)
        if qs.count() == 0:
            raise PPA.DoesNotExist('No free PPAs exist in pool')
        pk = qs[0].pk
        qs.filter(pk__in=(pk,)).update(reserved=True, lockname=lockname)

        # we now hold a lock, be careful
        try:
            ppa = PPA.objects.get(pk=pk)
            ppa.timestamp = datetime.datetime.now()
            if clean:
                ppa.clean_artifacts()

            ppa.save()
            return ppa
        except Exception as e:
            if ppa:
                ppa.reserved = False
                ppa.lockname = ''
                ppa.save()
            if not isinstance(e, LaunchpadError):
                logging.exception('Unexpected exception while cleaning')
                e = LaunchpadError(e)
            raise e
