# Ubuntu CI Services
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import datetime
import json
import threading
import time
import urllib2

import mock

from django.conf import settings
from django.db import connection, transaction
from django.db.utils import ConnectionHandler, DEFAULT_DB_ALIAS
from django.test import TestCase, TransactionTestCase, skipUnlessDBFeature

from ci_utils.tastypie.test import TastypieTestCase
from ppa_assigner.models import PPA, LaunchpadError


class TestModel(TestCase):

    def setUp(self):
        super(TestModel, self).setUp()
        for n in range(5):
            PPA.objects.create(name='ci_pool-%s' % str(n).zfill(3))

    @mock.patch('ppa_assigner.models.PPA.clean_artifacts')
    def testReserveNoClean(self, clean):
        p = PPA.reserve('testlockname')
        self.assertIsNotNone(p)
        self.assertFalse(clean.called)
        self.assertEqual('testlockname', p.lockname)

        reserved = PPA.objects.filter(reserved=True).count()
        self.assertEqual(1, reserved)

    def testReserveTimestamp(self):
        p = PPA.reserve()
        delta = datetime.datetime.now() - p.timestamp
        # the lock should have happened in less than a second
        self.assertLess(delta.seconds, 1)

    @mock.patch('ppa_assigner.models.PPA.clean_artifacts')
    def testReserveClean(self, clean):
        p = PPA.reserve('testlockname', True)
        self.assertIsNotNone(p)
        self.assertTrue(clean.called)
        self.assertEqual('testlockname', p.lockname)

        reserved = PPA.objects.filter(reserved=True).count()
        self.assertEqual(1, reserved)

    @mock.patch('ppa_assigner.models.PPA.clean_artifacts')
    def testReserveCleanFail(self, clean):
        '''ensure ppa isn't locked if clean operation fails.'''
        clean.side_effect = RuntimeError
        locked_before = PPA.objects.filter(reserved=True).count()
        with self.assertRaises(LaunchpadError):
            PPA.reserve(clean=True)

        locked_after = PPA.objects.filter(reserved=True).count()
        self.assertEqual(locked_before, locked_after)

    def testReserveNoFree(self):
        PPA.objects.update(reserved=True)
        with self.assertRaises(PPA.DoesNotExist):
            PPA.reserve()

    def testCleanOperation(self):
        # TODO more logic for this
        PPA.reserve(clean=True)


class TestConcurrency(TransactionTestCase):
    # inspired by, but heavily modified:
    # https://github.com/django/django/blob/master/tests/select_for_update/

    def setUp(self):
        super(TestConcurrency, self).setUp()
        transaction.enter_transaction_management()
        for n in range(5):
            PPA.objects.create(name='ci_pool-%s' % str(n).zfill(3))
        transaction.commit()
        self.addCleanup(transaction.abort)

    @contextlib.contextmanager
    def blocking_transaction(self):
        # We need another database connection to test that one connection
        # issuing a SELECT ... FOR UPDATE will block.
        con = ConnectionHandler(settings.DATABASES)[DEFAULT_DB_ALIAS]
        con.enter_transaction_management()
        cursor = con.cursor()
        sql = 'SELECT * FROM %(db_table)s %(for_update)s;' % {
            'db_table': PPA._meta.db_table,
            'for_update': con.ops.for_update_sql(),
        }
        cursor.execute(sql, ())
        cursor.fetchone()
        try:
            yield
        finally:
            con.rollback()
            con.close()

    def _reserve(self, params):
        params['signal'].set()
        start = time.time()
        PPA.reserve()
        params['time'] = time.time() - start
        # connections are per-thread, so close this
        connection.close()

    @skipUnlessDBFeature('has_select_for_update')
    def testReserveConcurrency(self):
        params = {'signal': threading.Event()}
        t = threading.Thread(target=self._reserve, args=(params,))
        locktime = 2
        with self.blocking_transaction():
            t.start()
            # wait for thread to let us know its running
            params['signal'].wait()
            time.sleep(locktime)
        t.join()
        # it won't be exact but it should be pretty close
        self.assertLess(abs(locktime - params['time']), 0.05)


class TestApi(TastypieTestCase):

    def setUp(self):
        super(TestApi, self).setUp('/api/v1')
        for n in range(5):
            PPA.objects.create(name='ci_pool-%s' % str(n).zfill(3))

    def testGetAll(self):
        obj = self.get('ppa/')
        self.assertEqual(PPA.objects.all().count(), len(obj['objects']))

    def testGetQuery(self):
        p = PPA.reserve()
        obj = self.get('ppa/', {'reserved': False})
        count = PPA.objects.filter(reserved=False).count()
        self.assertEqual(count, len(obj['objects']))

        obj = self.getResource('ppa/', {'reserved': True})
        self.assertEqual(p.name, obj['name'])

        p.lockname = 'foo'
        p.save()
        obj = self.getResource('ppa/', {'reserved': True})
        self.assertEqual(p.lockname, obj['lockname'])

    def testReserveNoClean(self):
        params = {'lockname': 'foo'}
        obj = self.createResource('ppa/', params)
        self.assertEqual(obj['lockname'], params['lockname'])
        self.assertTrue(obj['reserved'])

    @mock.patch('ppa_assigner.models.PPA.clean_artifacts')
    def testReserveClean(self, clean):
        obj = self.createResource('ppa/', {'lockname': 'foo', 'clean': True})
        self.assertTrue(obj['reserved'])
        self.assertTrue(clean.called)

    def testReserveNoFree(self):
        PPA.objects.update(reserved=True)
        params = {'lockname': 'foo'}
        with self.assertRaises(PPA.DoesNotExist):
            self.post('ppa/', params)

    def testFree(self):
        p = PPA.reserve()
        self.patch('ppa/%d/' % p.id, {'reserved': False})
        p = PPA.objects.get(pk=p.id)
        self.assertFalse(p.reserved)

    @mock.patch('urllib2.urlopen')
    def testPopulateFail(self, urlopen):
        '''ensure a lp error is handled correctly.'''
        def side_effect(arg):
            raise urllib2.URLError('foo')
        urlopen.side_effect = side_effect
        resp = self.client.patch(
            self._resource('ppa/'),
            data={'populate': True},
            authentication=self.auth)
        self.assertEqual(500, resp.status_code)
        self.assertEqual('<urlopen error foo>', resp.content)

    @mock.patch('urllib2.urlopen')
    def testPopulateMerge(self, urlopen):
        data = {
            'total_size': 2,
            'entries': [
                {'name': 'ppa1'},
                {'name': 'ppa2'},
            ]
        }

        resp = mock.Mock()
        resp.read.return_value = json.dumps(data)
        urlopen.return_value = resp
        user = settings.LAUNCHPAD_PPA_USER
        self.patch('ppa/', {'populate': True})
        expected = ['ppa:%s/%s' % (user, x['name']) for x in data['entries']]
        found = [x.name for x in PPA.objects.all().order_by('name')]
        self.assertListEqual(expected, found)
