# Ubuntu Continuous Integration Engine
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.test import TestCase
from ci_utils.tastypie.test import TastypieTestCase
from project.models import SourcePackage, BinaryPackage


def create_sourcepackage(name="awesome-package"):
    package = SourcePackage()
    package.name = name
    package.save()
    return package


def create_binarypackage(name="awesome-binary", sourcepackage=None,
                         seeded=False):
    binary = BinaryPackage()
    binary.name = name
    binary.sourcepackage = sourcepackage
    binary.seeded = seeded
    binary.save()
    return binary


class SourcePackageModelTest(TestCase):

    def test_creating_a_sourcepackage(self):
        package = create_sourcepackage()
        package_in_database = SourcePackage.objects.all()
        self.assertEquals(len(package_in_database), 1)
        only_package_in_database = package_in_database[0]
        self.assertEquals(only_package_in_database, package)

        self.assertEquals(only_package_in_database.name, "awesome-package")


class BinaryPackageModelTest(TestCase):

    def setUp(self):
        self.sourcepackage = create_sourcepackage()
        self.binarypackage = create_binarypackage(
            sourcepackage=self.sourcepackage)

    def test_creating_a_binarypackage(self):
        binarypackage_in_database = BinaryPackage.objects.all()
        self.assertEquals(len(binarypackage_in_database), 1)
        only_binarypackage_in_database = binarypackage_in_database[0]
        self.assertEquals(only_binarypackage_in_database, self.binarypackage)

        self.assertEquals(only_binarypackage_in_database.name,
                          "awesome-binary")


class SourcePackageResourceTest(TastypieTestCase):

    def setUp(self):
        super(SourcePackageResourceTest, self).setUp('/api/v1')
        create_sourcepackage()
        self.sourcepackage_1 = SourcePackage.objects.get(
            name='awesome-package')
        self.detail_url = 'sourcepackage/{0}/'.format(self.sourcepackage_1.pk)

    def test_get_sourcepackage_list_json(self):
        obj = self.getResource('sourcepackage/')
        self.assertEqual(obj['objects'][0], {
            u'id': self.sourcepackage_1.pk,
            u'name': u'awesome-package',
            u'resource_uri': u'/api/v1/sourcepackage/{0}/'.format(
                self.sourcepackage_1.pk)
        })

    def test_get_sourcepackage_detail_json(self):
        obj = self.getResource(self.detail_url)
        # We use ``assertKeys`` here to just verify the keys, not all the data.
        self.assertKeys(obj, ['id', 'name', 'resource_uri'])
        self.assertEqual(obj['name'], 'awesome-package')


class BinaryPackageResourceTest(TastypieTestCase):

    def setUp(self):
        super(BinaryPackageResourceTest, self).setUp('/api/v1')
        create_sourcepackage()
        self.sourcepackage_1 = SourcePackage.objects.get(
            name='awesome-package')
        create_binarypackage(sourcepackage=self.sourcepackage_1)
        self.binarypackage_1 = BinaryPackage.objects.get(name='awesome-binary')
        self.detail_url = 'binarypackage/{0}/'.format(self.binarypackage_1.pk)

    def test_get_binarypackage_list_json(self):
        obj = self.getResource('binarypackage/')
        self.assertEqual(obj['objects'][0], {
            u'id': self.binarypackage_1.pk,
            u'name': u'awesome-binary',
            u'seeded': False,
            u'sourcepackage': {
                u'id': self.sourcepackage_1.pk,
                u'name': u'awesome-package',
                u'resource_uri': u'/api/v1/sourcepackage/{0}/'.format(
                    self.sourcepackage_1.pk)},
            u'resource_uri': u'/api/v1/binarypackage/{0}/'.format(
                self.binarypackage_1.pk)
        })

    def test_get_binarypackage_detail_json(self):
        obj = self.getResource(self.detail_url)

        # We use ``assertKeys`` here to just verify the keys, not all the data.
        self.assertKeys(obj, ['id', 'name', 'resource_uri', 'seeded',
                        'sourcepackage'])
        self.assertEqual(obj['name'], 'awesome-binary')
